using System;
using System.Runtime.InteropServices;

namespace GpioDotNet
{
    enum GpioEdgeSettings { GPIO_REACT_ON_FALLING, GPIO_REACT_ON_RISING, GPIO_REACT_ON_BOTH, GPIO_DONT_REACT };

    class GpioCmX300 : IGpioInterface
    {
        private IntPtr hDrv = IntPtr.Zero;
 
        // Device CreateFile specific constants
        internal const Int32 INVALID_HANDLE_VALUE = -1;

        internal uint IOCTL_GPIO_SET_INTERRUPT = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0826);
        internal uint IOCTL_GPIO_RELEASE_INTERRUPT = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0827);
        internal uint IOCTL_GPIO_SET_VALUE = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0828);
        internal uint IOCTL_GPIO_GET_VALUE = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0829);
        internal uint IOCTL_GPIO_INTERRUPT_DONE = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x082A);
        


        [StructLayout(LayoutKind.Sequential)]
        internal class IoctlGpioState
        {
            public IoctlGpioState(uint gpio, GpioConfiguration direction,GpioLevel level)
            {
                this.num = gpio;
                this.dir = (uint)((direction == GpioConfiguration.OUTPUT) ? 1 : 0);
                this.value = (uint)level;
            }
            public IoctlGpioState(uint gpio, GpioConfiguration direction)
            {
                this.num = gpio;
                this.dir = (uint)((direction == GpioConfiguration.OUTPUT) ? 1 : 0);
                this.value = 0 ;
            }
            public uint num;
            public uint dir;
            public uint value;
        }


        [StructLayout(LayoutKind.Sequential)]
        internal class IoctlGpioInterrupt
        {
            public IoctlGpioInterrupt(uint gpio, GpioEdgeSettings edgeSetting, IntPtr eventHandle)
            {
                this.num = gpio;
                this.edge = (uint)edgeSetting;
                this.hEvent = (uint)eventHandle.ToInt32();
            }
            public uint num;
            public uint edge;
            public uint hEvent;
        }


        private bool SendIOCTL(uint ioctl, Object inputData, uint inputParametersSize)
        {
            UInt32 stam;
            return SendIOCTL(ioctl, inputData, inputParametersSize, out stam, 0);
        }

        private UInt32 GetIOCTL(uint ioctl, uint value)
        {
            UInt32 gpioNumber = value;
            UInt32 retVal = 0;

            SendIOCTL(ioctl, gpioNumber, (uint)Marshal.SizeOf(gpioNumber), out retVal, (uint)Marshal.SizeOf(retVal));
            return retVal;
        }

        private bool SendIOCTL(uint ioctl, Object inputData, uint inputParametersSize, out UInt32 outputData, uint outputParametersSize)
        {
            uint bytesReturned = 0;
            // Initialize unmanged memory to hold the struct.
            IntPtr inputParametersPtr = IntPtr.Zero;
            IntPtr outputParametersPtr = IntPtr.Zero;
            try
            {
                if (inputParametersSize > 0)
                {
                    inputParametersPtr = Marshal.AllocHGlobal((int)inputParametersSize);
                    if (inputParametersSize > 4)
                        Marshal.StructureToPtr(inputData, inputParametersPtr, false);
                    else
                        Marshal.WriteInt32(inputParametersPtr,Convert.ToInt32(inputData));
                            
                }
                if(outputParametersSize > 0)
                    outputParametersPtr = Marshal.AllocHGlobal((int)outputParametersSize);
                // Copy the struct to unmanaged memory.
                
                bool res = NativeMethods.DeviceIoControl(hDrv, ioctl,
                                           inputParametersPtr, inputParametersSize,
                                           outputParametersPtr, outputParametersSize,
                                           ref bytesReturned, IntPtr.Zero);
                if (!res)
                {
                    // Throw Win32 exception with error code
                    int err = Marshal.GetLastWin32Error();
                    throw new System.ComponentModel.Win32Exception(err);
                }
                if (outputParametersSize > 0)
                {
                    outputData = (UInt32)Marshal.ReadInt32(outputParametersPtr);
                }
                else
                    outputData = 0;
            }
            catch (Exception ex)
            {
                // throw all other exceptions to upper levels
                throw ex;
            }
            finally
            {
                if(inputParametersSize > 0)
                    Marshal.FreeHGlobal(inputParametersPtr);
                if(outputParametersSize > 0)
                    Marshal.FreeHGlobal(outputParametersPtr);
            }
            return true;
        }

        
        public GpioCmX300() 
        {
            //First, we create a handle to the Gpio driver. GIO1: in our case.
            hDrv = NativeMethods.CreateFile("GIO1:", 0, 0, IntPtr.Zero, NativeMethods.OPEN_ALWAYS, 0, IntPtr.Zero);
            if (hDrv.ToInt32() == INVALID_HANDLE_VALUE)
                throw new InvalidOperationException("Can't open GIO1:");
        }

        ~GpioCmX300()
        {
            NativeMethods.CloseHandle(hDrv);
        }

        public bool SetAsOutput(uint gpioNumber, GpioLevel level)
        {
            IoctlGpioState inputParameters = new IoctlGpioState(gpioNumber,GpioConfiguration.OUTPUT,level);
            return SendIOCTL(IOCTL_GPIO_SET_VALUE, inputParameters, (uint)Marshal.SizeOf(inputParameters));
        }

        public bool SetAsInput(uint gpioNumber)
        {
            IoctlGpioState inputParameters = new IoctlGpioState(gpioNumber, GpioConfiguration.INPUT);
            return SendIOCTL(IOCTL_GPIO_SET_VALUE, inputParameters, (uint)Marshal.SizeOf(inputParameters));
        }

       
        /**
         *  Not supported for CM-X300!
        */
        public GpioConfiguration GetConfiguration(uint gpioNumber)
        {
            return GpioConfiguration.OUTPUT;
        }

        /**
        // Get current level of an output gpio
        */
        public GpioLevel GetGpioLevel(uint gpioNumber)
        {
            UInt32 uiRes = GetIOCTL(IOCTL_GPIO_GET_VALUE, gpioNumber);
            if (uiRes != 0)
                return GpioLevel.HIGH;
            else
                return GpioLevel.LOW;
        }

        // Only works for OMAP GPIOs (not TWL GPIOs)
        public bool SetupInterrupt(uint gpioNumber, bool triggerOnFallingEdge, bool triggerOnRisingEdge)
        {
            GpioEdgeSettings edge = GpioEdgeSettings.GPIO_DONT_REACT;
            if (triggerOnFallingEdge) 
            {
                if (triggerOnRisingEdge)
                    edge = GpioEdgeSettings.GPIO_REACT_ON_BOTH;
                else
                    edge = GpioEdgeSettings.GPIO_REACT_ON_FALLING;
            }
            else if (triggerOnRisingEdge) 
                edge = GpioEdgeSettings.GPIO_REACT_ON_RISING;

            IoctlGpioInterrupt inputParameters = new IoctlGpioInterrupt(gpioNumber, edge, IntPtr.Zero);
            return SendIOCTL(IOCTL_GPIO_SET_INTERRUPT, inputParameters, (uint)Marshal.SizeOf(inputParameters));
        }
        
        public bool AckInterrupt(uint gpioNumber)
        {
            return SendIOCTL(IOCTL_GPIO_INTERRUPT_DONE, gpioNumber, (uint)Marshal.SizeOf(gpioNumber));
        }

        public bool DisableInterrupt(uint gpioNumber)
        {
            return SendIOCTL(IOCTL_GPIO_RELEASE_INTERRUPT, gpioNumber, (uint)Marshal.SizeOf(gpioNumber));
        }
    }
}