using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text;

namespace I2CDotNet
{
    enum Pxa_I2c_Speed
    {
        PXA_I2C_FAST_SPEED = 1,
        PXA_I2C_NORMAL_SPEED
    }

    enum Pxa_I2c_Operation_Code
    {
	    I2C_OPCODE_READ	=	1,
	    I2C_OPCODE_WRITE=	2
    } 

    class I2C_CmX300 : I2CDriverInterface
    {
        private IntPtr ipI2Chandle = IntPtr.Zero;
        private byte slaveAddr;
        
        // Device CreateFile specific constants
        internal const Int32 INVALID_HANDLE_VALUE = -1;
        internal const UInt32 CREATE_NEW = 1;
        internal const UInt32 OPEN_ALWAYS = 4;
        //internal const UInt32 GENERIC_READ = 0x80000000;
        //internal const UInt32 GENERIC_WRITE = 0x40000000;

        internal uint IOCTL_I2C_TRANSACT = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0810);

        [StructLayout(LayoutKind.Sequential)]
        internal unsafe class Pxa_I2c_Trans
        {
            public Pxa_I2c_Trans(uint transactions,Pxa_I2c_Speed speed,byte addr, Pxa_I2c_Operation_Code* operation,uint* offset,uint* len, byte* bufferPtr)
            {
                this.numOfTransactions=transactions;
                this.deviceAddr=addr;
                this.opCode=operation;
                this.speedCode=speed;
                this.bufferOffset=offset;
                this.transLen=len;
                this.buffer=bufferPtr;
            }
            public uint numOfTransactions;
            public Pxa_I2c_Speed speedCode;
            public byte deviceAddr;
            public Pxa_I2c_Operation_Code* opCode;
            public uint* bufferOffset;
            public uint* transLen;
            public byte* buffer;
        }
 

        public I2C_CmX300(int port)
        {
            string driverName = "I2C" + port + ":";
            //First, we create a handle to the I2C proxy driver. I2C3 in our case.
            ipI2Chandle = NativeMethods.CreateFile(driverName, 0, 0, IntPtr.Zero,( OPEN_ALWAYS | CREATE_NEW), 0, IntPtr.Zero);
            if (ipI2Chandle.ToInt32() == INVALID_HANDLE_VALUE)
                throw new InvalidOperationException("Can't open " + driverName);
        }

        ~I2C_CmX300()
        {
            NativeMethods.CloseHandle(ipI2Chandle);
        }


        public unsafe bool I2CUserTransact(uint transactions, Pxa_I2c_Speed speed, Pxa_I2c_Operation_Code[] operation, uint[] offset, uint[] len,ref byte[] bufferPtr)
        {
            fixed (Pxa_I2c_Operation_Code* opBuffFixed = operation)
            {
                fixed (uint* offsetBuffFixed = offset)
                {
                    fixed (uint* lenBuffFixed = len)
                    {
                        fixed (byte* outBuffFixed = bufferPtr)
                        {

                            Pxa_I2c_Trans transaction = new Pxa_I2c_Trans(transactions, speed, slaveAddr, opBuffFixed, offsetBuffFixed, lenBuffFixed, outBuffFixed);
                            uint bytesReturned = 0;
                            // Initialize unmanged memory to hold the struct.
                            IntPtr inputParametersPtr = IntPtr.Zero;
                            IntPtr outputParametersPtr = IntPtr.Zero;
                            uint inputParametersSize = 0;
                            uint outputParametersSize = 0;
                            try
                            {
                                outputParametersSize = (uint)Marshal.SizeOf(transaction);
                                outputParametersPtr = Marshal.AllocHGlobal((int)outputParametersSize);
                                // Copy the struct to unmanaged memory.
                                Marshal.StructureToPtr(transaction, outputParametersPtr, false);
                                bool res = NativeMethods.DeviceIoControl(ipI2Chandle, IOCTL_I2C_TRANSACT,
                                                           inputParametersPtr, inputParametersSize,
                                                           outputParametersPtr, outputParametersSize,
                                                           ref bytesReturned, IntPtr.Zero);
                                Marshal.PtrToStructure(outputParametersPtr, transaction);
                                Marshal.FreeHGlobal(outputParametersPtr);
                                if (!res)
                                {
                                    // Throw Win32 exception with error code
                                    int err = Marshal.GetLastWin32Error();
                                    throw new System.ComponentModel.Win32Exception(err);
                                }
                            }
                            catch (Exception ex)
                            {
                                // throw all other exceptions to upper levels
                                throw ex;
                            }
                        }
                    }
                }
            }
            return true;
        }

        public bool SetSlaveAddress(uint slaveAddress)
        {
            slaveAddr = (byte)slaveAddress;
            return true;
        }
        

         /// <summary>
        /// Read from I2C device
        /// </summary>
        /// <param name="subAddress">Device internal address</param>
        /// <param name="buffer"> [out] The data will be written to this array</param>
        /// <param name="size"> number of bytes to read </param>
        /// <returns></returns>
        public bool Read(uint subAddress, byte[] buffer, uint size)
        {
            Pxa_I2c_Operation_Code[] OpCode = new Pxa_I2c_Operation_Code[2];
            uint[] BufferOffset = new uint[2];
            uint[] TransLen = new uint[2];
            byte[] DataBuffer = new byte[size + 1];
       
            OpCode[0] = Pxa_I2c_Operation_Code.I2C_OPCODE_WRITE;
            BufferOffset[0] = 0;
            TransLen[0] = 1;
            DataBuffer[0] =(byte)subAddress;

            //2nd transaction is read register contents
            OpCode[1] = Pxa_I2c_Operation_Code.I2C_OPCODE_READ;
            BufferOffset[1] = 1;
            TransLen[1] = 2;
            DataBuffer[1] = (byte)subAddress;

            I2CUserTransact(2, Pxa_I2c_Speed.PXA_I2C_FAST_SPEED, OpCode, BufferOffset, TransLen, ref DataBuffer);

            for(int i=1;i<(size+1);i++)
                buffer[i-1] = DataBuffer[i];

            return true;
        }

        /// <summary>
        /// Write to I2C device
        /// </summary>
        /// <param name="subAddress"> Device internal address</param>
        /// <param name="buffer"> data to be written to the device </param>
        /// <param name="size"> number of bytes to write</param>
        /// <returns></returns>
        public bool Write(uint subAddress, byte[] buffer, uint size)
        {
            Pxa_I2c_Operation_Code[] OpCode = new Pxa_I2c_Operation_Code[2];
            uint[] BufferOffset = new uint[2];
            uint[] TransLen = new uint[2];
            byte[] DataBuffer = new byte[size + 1];
            
            OpCode[0] = Pxa_I2c_Operation_Code.I2C_OPCODE_WRITE;
            BufferOffset[0] = 0;
            TransLen[0] = 3;
            DataBuffer[0] = (byte)subAddress;

            for (int i = 1; i < (size+1); i++)
                DataBuffer[i ]= buffer[i-1];

            I2CUserTransact(1, Pxa_I2c_Speed.PXA_I2C_FAST_SPEED, OpCode, BufferOffset, TransLen, ref DataBuffer);
            return true;
        }
    }
}
