using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace I2CDotNet
{
    /**********************************************************************************************************
     About I2CDotNet Example.
    ***********************************************************************************************************
     This example shows how to transact with the CAT9555 Gpio extender. The extender communicates thrgough the
     I2C bus.
     All register are set to 0xFF after reset. Another feature we use in this device is the pairing of registers,
     meaning that if you address say Configuration register number 1 and write 2 bytes, then the 1st byte is written
     to configuration register 1 and the 2nd byte is written to configuration register 0.
     In this example we do the following:
     -Read Configuration registers (to see value before setting it)
     -Writing to Configuration registers
     -Read Configuration registers (to confirm that the write was successful - just for this example)
     -Read the Output Port registers (to see value before setting it)
     -Write the Output Port registers
     -Read Output Port registers (to confirm that the write was successful - just for this example)

     When done you will see that all of the pins, except for pin 1, have a logical value 1 (the default value)
     and pin #1 outputs a logical value of 0 (you can use any measuring device you like)
    ************************************************************************************************************/

    class Program
    {
        //Set values for this sample
        internal const byte SLAVE_ADDRESS_EXTENDER	= 0x24;	//0b0100100
        internal const byte REGISTER_CONFIG0	        = 0x6;
        internal const byte REGISTER_OUTPORT          = 0x2;

        private static I2C_CmX300 hI2C;

        static void Main(string[] args)
        {
            /* Allocating memory for the buffers - a transaction can be of the size you desire */
            Pxa_I2c_Operation_Code[]	OpCode= new Pxa_I2c_Operation_Code[5];
            uint[]	BufferOffset=new uint[5];
            uint[]	TransLen=new uint[5];
            byte[]	Buffer=new byte[5];

            uint numOfTransactions = 0;

            //Creates an instance of the gpio driver .Net wrapper
            hI2C = new I2C_CmX300(1);
            
            //Choose slave address
            //The PXA-300 only supports 7 bit slave address so make sure 
            //the address you have chosen isn't greater than 127...
            hI2C.SetSlaveAddress(SLAVE_ADDRESS_EXTENDER);

            //Read the configuration register to see that after we will write it
            //It will change
            numOfTransactions = 2;
        
            //We first set the address of the register we want to read from
            //by writing it's address
            
            //1st transaction is choose register by writing it's address to the device
            OpCode[0] = Pxa_I2c_Operation_Code.I2C_OPCODE_WRITE;
            BufferOffset[0] = 0;
            TransLen[0] = 1;
            Buffer[0] = REGISTER_CONFIG0;

            //2nd transaction is read register contents
            OpCode[1] = Pxa_I2c_Operation_Code.I2C_OPCODE_READ;
            BufferOffset[1] = 1;
            TransLen[1] = 2;
            Buffer[1] = REGISTER_CONFIG0;

            //Using RegulareMode clock just for this transaction
            hI2C.I2CUserTransact(numOfTransactions, Pxa_I2c_Speed.PXA_I2C_NORMAL_SPEED, OpCode, BufferOffset, TransLen, ref Buffer);

            System.Console.WriteLine("Read Configuration before writing to it: High=" + Buffer[1] + ", Low=" + Buffer[2]);

            //Configure GPIO 0.0 to be Output

            //Writing to the register is done in one transaction
            numOfTransactions = 1;
            OpCode[0] = Pxa_I2c_Operation_Code.I2C_OPCODE_WRITE;
            BufferOffset[0] = 0;
            TransLen[0] = 3;
            Buffer[0] = REGISTER_CONFIG0;
            Buffer[1] = 0x02;//Setting pin 0.0 as output and all the rest as inputs
            Buffer[2] = 0x00;

            hI2C.I2CUserTransact(numOfTransactions, Pxa_I2c_Speed.PXA_I2C_FAST_SPEED, OpCode, BufferOffset, TransLen, ref Buffer);

            //Read the configuration register to see that after we will write it
            //It will change
            numOfTransactions = 2;

            //We first set the address of the register we want to read from
            //by writing it's address

            //First transaction is choose register
            OpCode[0] = Pxa_I2c_Operation_Code.I2C_OPCODE_WRITE;
            BufferOffset[0] = 0;
            TransLen[0] = 1;
            Buffer[0] = REGISTER_CONFIG0;


            OpCode[1] = Pxa_I2c_Operation_Code.I2C_OPCODE_READ;
            BufferOffset[1] = 1;
            TransLen[1] = 2;

            hI2C.I2CUserTransact(numOfTransactions, Pxa_I2c_Speed.PXA_I2C_FAST_SPEED, OpCode, BufferOffset, TransLen, ref Buffer);

            System.Console.WriteLine("Read Configuration After writing to it: High=" + Buffer[1] + ", Low=" + Buffer[2]);
            
            //Read the OUTPUTPORT register to see that after we will write it
            //It will change
            numOfTransactions = 2;

            //We first set the address of the register we want to read from
            //by writing it's address

            //First transaction is choose register
            OpCode[0] = Pxa_I2c_Operation_Code.I2C_OPCODE_WRITE;
            BufferOffset[0] = 0;
            TransLen[0] = 1;
            Buffer[0] = REGISTER_OUTPORT;
            
            OpCode[1] = Pxa_I2c_Operation_Code.I2C_OPCODE_READ;
            BufferOffset[1] = 1;
            TransLen[1] = 2;

            hI2C.I2CUserTransact(numOfTransactions, Pxa_I2c_Speed.PXA_I2C_FAST_SPEED, OpCode, BufferOffset, TransLen, ref Buffer);

            System.Console.WriteLine("Read output port Before writing to it: High=" + Buffer[1] + ", Low=" + Buffer[2]);

            //Again writing is done in 1 transaction
            numOfTransactions = 1;

            //First transaction is choose register
            OpCode[0] = Pxa_I2c_Operation_Code.I2C_OPCODE_WRITE;
            BufferOffset[0] = 0;
            TransLen[0] = 3;
            Buffer[0] = REGISTER_OUTPORT;
            Buffer[1] = 0xfe; //Pin 0.0 is output, setting it to Low.
            Buffer[2] = 0xff;

            hI2C.I2CUserTransact(numOfTransactions, Pxa_I2c_Speed.PXA_I2C_FAST_SPEED, OpCode, BufferOffset, TransLen, ref Buffer);

            //Let's read the Output port register to see that writing to it changed things
            //(It defaults to 0xFF)
            numOfTransactions = 2;
            
            //We first set the address of the register we want to read from
            //by writing it's address

            //First transaction is choose register
            OpCode[0] = Pxa_I2c_Operation_Code.I2C_OPCODE_WRITE;
            BufferOffset[0] = 0;
            TransLen[0] = 1;
            Buffer[0] = REGISTER_OUTPORT;

            //Now we read the contents
            OpCode[1] = Pxa_I2c_Operation_Code.I2C_OPCODE_READ;
            BufferOffset[1] = 0;
            TransLen[1] = 2;

            hI2C.I2CUserTransact(numOfTransactions, Pxa_I2c_Speed.PXA_I2C_FAST_SPEED, OpCode, BufferOffset, TransLen, ref Buffer);

           
            //printf("Read Output Port rgisters: OutPort0: 0x%02x, OutPort1: 0x%02x\r\n", trans.mBuffer[0], trans.mBuffer[1]);
            //printf("You can now check to see that the pin is high\r\n");

            System.Console.WriteLine("Read output port registers: High=" + Buffer[1] + ", Low=" + Buffer[2]);
            System.Console.WriteLine("You can now check to see that the pin is low.");

            Thread.Sleep(1000);

            byte[] Buffer2 = new byte[2];

            Buffer2[0] = 0xff;
            Buffer2[1] = 0xff;

            hI2C.Write(REGISTER_OUTPORT, Buffer2, 2);

            Buffer2[0] = 0;
            Buffer2[1] = 0;
            
            hI2C.Read(REGISTER_OUTPORT, Buffer2, 2);

            System.Console.WriteLine("Read output port registers2: High=" + Buffer2[0] + ", Low=" + Buffer2[1]);
        }
    }
}
