using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace SpiDotNet
{
    enum Polarity { LOW_INACTIVE, HIGH_INACTIVE };

    class SpiCmX300 : ISpiInterface
    {
        private IntPtr hDrv = IntPtr.Zero;

        internal const uint SSP3_FREQ = 4;	//Specified in KHz, accroding to spec needs to be between DC and 5MHz
        internal const uint SSP3_DATA_LEN = 8;	//Each transaction is 8 bits (writing 8 and reading 8 minimum)

        // Device CreateFile specific constants
        internal const Int32 INVALID_HANDLE_VALUE = -1;
        internal const UInt32 CREATE_NEW  = 1;
        internal const UInt32 OPEN_ALWAYS = 4;

        internal uint IOCTL_SPI_TRANSACT               = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0840);
        internal uint IOCTL_SPI_REGISTER_CONFIGURATION = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0842);

        [StructLayout(LayoutKind.Sequential)]
        internal class IoctlSpiSlaveConfig
        {
            public IoctlSpiSlaveConfig(uint DataSize, uint ClkRate, 
                Polarity ClkPol,Polarity ChipSelPol,bool TxTristate,bool ClkPhaseHalfCycle,bool ChipSelExtrn,uint GpioNum)
            {
                this.dwDataSize = DataSize;
                this.dwClockRate = ClkRate;
                this.ClockPol = ClkPol;
                this.ChipSelectPol = ChipSelPol;
                this.bTxTristate = (uint)(TxTristate ? 1 : 0);
                this.ClockPhaseHalfCycle = (uint)(ClkPhaseHalfCycle ? 1 : 0);
                this.bCSExternal = (uint)(ChipSelExtrn ? 1 : 0);
                this.dwGpioNum = GpioNum;
            }
            public uint dwDataSize;          //must be in the range of [4,32]
	        public uint	dwClockRate;         //in KHz, must be in the range of [4, 13*1024]
            public Polarity ClockPol;
            public Polarity ChipSelectPol;
            public uint bTxTristate;
            public uint ClockPhaseHalfCycle;
            public uint bCSExternal;         //Is Chip Select a standard GPIO?
	        public uint dwGpioNum;           //Valid if bCSExternal is TRUE
        }

        [StructLayout(LayoutKind.Sequential)]
        internal unsafe class IoctlSpiTrans
        {
            public IoctlSpiTrans(uint numOfTransactions ,uint* writeBuffPtr ,uint* readBuffPtr)
            {
                this.mTransactions = numOfTransactions;
                this.mWriteBuffer  = writeBuffPtr;
                this.mReadBuffer   = readBuffPtr;
            }
            public uint mTransactions;
            public uint* mWriteBuffer;
            public uint* mReadBuffer;
        }

        public SpiCmX300(uint num) 
        {
            string driverName = "SPI" + num + ":";
            //First, we create a handle to the SPI driver.
            hDrv = NativeMethods.CreateFile(driverName, 0, 0, IntPtr.Zero, (OPEN_ALWAYS | CREATE_NEW), 0, IntPtr.Zero);
            if (hDrv.ToInt32() == INVALID_HANDLE_VALUE)
                throw new InvalidOperationException("Can't open " + driverName);
        }

        ~SpiCmX300()
        {
            NativeMethods.CloseHandle(hDrv);
        }
        public bool SPIConfigure(uint address, uint config) 
        {
            Polarity ClkPol = (config & (1 << 1)) == (1 << 1) ? Polarity.HIGH_INACTIVE : Polarity.LOW_INACTIVE;
            Polarity ChipSelPol = (config & (1 << 6)) == (1 << 6) ? Polarity.HIGH_INACTIVE : Polarity.LOW_INACTIVE;
            return SPIUserSetConfiguration(SSP3_DATA_LEN,
                                             SSP3_FREQ,
                                             ClkPol,
                                             ChipSelPol, // Polarity of CS
                                             true,  // Will TX line be tristated when CS is deasserted
                                             true,
                                             false,
                                             0);    // Will the SSP port drive the Chip Select or is it an external GPIO we use - Here we use the internal controller
        }
        public bool SPIWriteRead(uint size, ref uint[] pOutBuffer, uint[] pInBuffer)
        {
            return SPIUserTransact(size, pInBuffer, ref pOutBuffer);
        }
 
        public bool SPIUserSetConfiguration(uint DataSize, uint ClkRate,
                Polarity ClkPol, Polarity ChipSelPol, bool TxTristate, bool ClkPhaseHalfCycle, bool ChipSelExtrn, uint GpioNum)
        {
            IoctlSpiSlaveConfig conf = new IoctlSpiSlaveConfig(DataSize, ClkRate, ClkPol, ChipSelPol, TxTristate,
                                                                ClkPhaseHalfCycle, ChipSelExtrn, GpioNum);
            return NativeMethods.SendIOCTL(hDrv, IOCTL_SPI_REGISTER_CONFIGURATION, conf, (uint)Marshal.SizeOf(conf));
        }

        public unsafe bool SPIUserTransact(uint numOfTransactions, uint[] writeBuff, ref uint[] readBuff) 
        {
            fixed (uint* inBuffFixed = writeBuff)
            {
                fixed (uint* outBuffFixed = readBuff)
                {

                    IoctlSpiTrans transaction = new IoctlSpiTrans(numOfTransactions, inBuffFixed, outBuffFixed);
                    uint bytesReturned = 0;
                    // Initialize unmanged memory to hold the struct.
                    IntPtr inputParametersPtr = IntPtr.Zero;
                    IntPtr outputParametersPtr = IntPtr.Zero;
                    uint inputParametersSize = 0;
                    uint outputParametersSize = 0;
                    try
                    {
                        outputParametersSize = (uint)Marshal.SizeOf(transaction);
                        outputParametersPtr = Marshal.AllocHGlobal((int)outputParametersSize);
                        // Copy the struct to unmanaged memory.
                        Marshal.StructureToPtr(transaction, outputParametersPtr, false);
                        bool res = NativeMethods.DeviceIoControl(hDrv, IOCTL_SPI_TRANSACT,
                                                   inputParametersPtr, inputParametersSize,
                                                   outputParametersPtr, outputParametersSize,
                                                   ref bytesReturned, IntPtr.Zero);
                        Marshal.PtrToStructure(outputParametersPtr, transaction);
                        Marshal.FreeHGlobal(outputParametersPtr);
                        if (!res)
                        {
                            // Throw Win32 exception with error code
                            int err = Marshal.GetLastWin32Error();
                            throw new System.ComponentModel.Win32Exception(err);
                        }
                    }
                    catch (Exception ex)
                    {
                        // throw all other exceptions to upper levels
                        throw ex;
                    }
                }
            }

            return true;
        }
    }
}
